/* 
 *  This program is an interactive FAQ to ease investigations.
 *  Copyright (C) 2012 Julien Bruguier.
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */ 

#ifndef __INVESTIGATION_DONNEES_REGEX__
#define __INVESTIGATION_DONNEES_REGEX__

#include <set>
#include <vector>
#include <src/global/global.h>
namespace Investigation { namespace Donnees { } }
#include <src/investigation/donnees/automate.h>

namespace Investigation
{
	namespace Donnees
	{
		typedef enum class { POSITIF, NEGATIF } TypeEnsembleLettre;
		template<typename Lettre> class ExpressionRationnelle;

		template<typename Lettre>
		struct Glouton
		{
			typedef Automate<NonDeterministeSimplifie,Lettre> AutomateEquivalent;
		};

		template<typename Lettre>
		struct NonGlouton
		{
			typedef Automate<Deterministe,Lettre> AutomateEquivalent;
		};

		template<template<typename Lettre> class Type, typename Lettre>
		class Patron
		{
			friend class ExpressionRationnelle<Lettre>;
			public:
				typedef std::vector<Lettre> Mot;
				bool operator () (const Mot& mot) const
				{
					return _automate(mot);
				};
				template<typename oStream>
				friend oStream& operator<<(oStream& os, const Patron<Type,Lettre>& patron)
				{
					os << patron._automate;
					return os;
				}
			private:
				Patron(const typename Type<Lettre>::AutomateEquivalent& automate)
				:_automate(automate) { };
				typename Type<Lettre>::AutomateEquivalent _automate;
		};

		template<typename Lettre>
		class ExpressionRationnelle
		{
			public:
				typedef std::set<Lettre> Alphabet;
				ExpressionRationnelle() { };
				virtual ~ExpressionRationnelle() {};
				template<template<typename _Lettre> class Type>
				operator Patron<Type,Lettre> () const
				{
					Alphabet alphabet = (*this);
					Automate<NonDeterministe,Lettre> automate = (*this)(alphabet);
					//std::cout << automate;
					return Patron<Type,Lettre>(automate);
				};
				virtual operator Alphabet () const = 0;
				virtual Automate<NonDeterministe,Lettre> operator()(const Alphabet& alphabet) const = 0;
			private:
		};

		template<typename Lettre>
		class ExpressionRationnelleLettre : public ExpressionRationnelle<Lettre>
		{
			public:
				explicit ExpressionRationnelleLettre(const Lettre& lettre)
				:_lettre(lettre) { };
				virtual ~ExpressionRationnelleLettre() {};
				virtual operator typename ExpressionRationnelle<Lettre>::Alphabet () const
				{
					typename ExpressionRationnelle<Lettre>::Alphabet alphabet;
					alphabet.insert(_lettre);
					return alphabet;
				};
				virtual Automate<NonDeterministe,Lettre> operator()(const typename ExpressionRationnelle<Lettre>::Alphabet& alphabet) const
				{
					Automate<NonDeterministe,Lettre> automate(alphabet);
					auto etatinitial = automate.ajout_etat_initial();
					auto etatfinal = automate.ajout_etat_final();
					automate.ajout_transition(etatinitial,_lettre,etatfinal);
					return automate;
				};
			private:
				Lettre _lettre;
		};

		template<typename Lettre, TypeEnsembleLettre type>
		class ExpressionRationnelleEnsembleLettre : public ExpressionRationnelle<Lettre>
		{
			public:
				explicit ExpressionRationnelleEnsembleLettre(const std::set<Lettre>& ensemble)
				:_ensemble(ensemble) { };
				virtual ~ExpressionRationnelleEnsembleLettre() {};
				virtual operator typename ExpressionRationnelle<Lettre>::Alphabet () const
				{
					return _ensemble;
				}
				virtual Automate<NonDeterministe,Lettre> operator()(const typename ExpressionRationnelle<Lettre>::Alphabet& alphabet) const
				{
					Automate<NonDeterministe,Lettre> automate(alphabet);
					auto etatinitial = automate.ajout_etat_initial();
					auto etatfinal = automate.ajout_etat_final();
					switch(type)
					{
						case TypeEnsembleLettre::POSITIF:
							{
								for(auto it=_ensemble.cbegin() ; it!=_ensemble.cend() ; ++it)
								{
									automate.ajout_transition(etatinitial,*it,etatfinal);
								}
							}
							break;
						case TypeEnsembleLettre::NEGATIF:
							{
								for(auto it=alphabet.cbegin() ; it!=alphabet.cend() ; ++it)
								{
									if(_ensemble.find(*it)==_ensemble.end())
									{
										automate.ajout_transition(etatinitial,*it,etatfinal);
									}
								}
								automate.ajout_transition(etatinitial,etatfinal);
							}
							break;
					}
					return automate;
				};
			private:
				std::set<Lettre> _ensemble;
		};

		template<typename Lettre>
		class ExpressionRationnelleJoker : public ExpressionRationnelle<Lettre>
		{
			public:
				explicit ExpressionRationnelleJoker() { };
				virtual ~ExpressionRationnelleJoker() {};
				virtual operator typename ExpressionRationnelle<Lettre>::Alphabet () const
				{	
					return typename ExpressionRationnelle<Lettre>::Alphabet();
				};
				virtual Automate<NonDeterministe,Lettre> operator()(const typename ExpressionRationnelle<Lettre>::Alphabet& alphabet) const
				{
					Automate<NonDeterministe,Lettre> automate(alphabet);
					auto etatinitial = automate.ajout_etat_initial();
					auto etatfinal = automate.ajout_etat_final();
					for(auto it=alphabet.cbegin() ; it!=alphabet.cend() ; ++it)
					{
						automate.ajout_transition(etatinitial,*it,etatfinal);
					}
					automate.ajout_transition(etatinitial,etatfinal);
					return automate;
				};
		};

		template<typename Lettre>
		class ExpressionRationnelleOptionnel : public ExpressionRationnelle<Lettre>
		{
			public:
				explicit ExpressionRationnelleOptionnel(const SHARED_PTR(ExpressionRationnelle<Lettre> )& expression)
				:_expression(expression) { };
				virtual ~ExpressionRationnelleOptionnel() {};
				virtual operator typename ExpressionRationnelle<Lettre>::Alphabet () const
				{	
					return (*_expression);
				};
				virtual Automate<NonDeterministe,Lettre> operator()(const typename ExpressionRationnelle<Lettre>::Alphabet& alphabet) const
				{
					Automate<NonDeterministe,Lettre> automate = (*_expression)(alphabet);
					automate.ajout_transition_instantanee(automate._etats_initiaux,automate._etats_finaux);
					return automate;
				};
			private:
				SHARED_PTR(ExpressionRationnelle<Lettre> ) _expression;
		};

		template<typename Lettre>
		class ExpressionRationnelleRepetitionOptionnelle : public ExpressionRationnelle<Lettre>
		{
			public:
				explicit ExpressionRationnelleRepetitionOptionnelle(const SHARED_PTR(ExpressionRationnelle<Lettre> )& expression)
				:_expression(expression) { };
				virtual ~ExpressionRationnelleRepetitionOptionnelle() {};
				virtual operator typename ExpressionRationnelle<Lettre>::Alphabet () const
				{	
					return (*_expression);
				};
				virtual Automate<NonDeterministe,Lettre> operator()(const typename ExpressionRationnelle<Lettre>::Alphabet& alphabet) const
				{
					Automate<NonDeterministe,Lettre> automate = (*_expression)(alphabet);
					automate.ajout_transition_instantanee(automate._etats_finaux,automate._etats_initiaux);
					automate.ajout_transition_instantanee(automate._etats_initiaux,automate._etats_finaux);
					return automate;
				};
			private:
				SHARED_PTR(ExpressionRationnelle<Lettre> ) _expression;
		};

		template<typename Lettre>
		class ExpressionRationnelleRepetitionObligatoire : public ExpressionRationnelle<Lettre>
		{
			public:
				explicit ExpressionRationnelleRepetitionObligatoire(const SHARED_PTR(ExpressionRationnelle<Lettre> )& expression)
				:_expression(expression) { };
				virtual ~ExpressionRationnelleRepetitionObligatoire() {};
				virtual operator typename ExpressionRationnelle<Lettre>::Alphabet () const
				{	
					return (*_expression);
				};
				virtual Automate<NonDeterministe,Lettre> operator()(const typename ExpressionRationnelle<Lettre>::Alphabet& alphabet) const
				{
					Automate<NonDeterministe,Lettre> automate = (*_expression)(alphabet);
					automate.ajout_transition_instantanee(automate._etats_finaux,automate._etats_initiaux);
					return automate;
				};
			private:
				SHARED_PTR(ExpressionRationnelle<Lettre> ) _expression;
		};

		template<typename Lettre>
		class ExpressionRationnelleSequence : public ExpressionRationnelle<Lettre>
		{
			public:
				explicit ExpressionRationnelleSequence() { };
				virtual ~ExpressionRationnelleSequence() {};
				ExpressionRationnelleSequence<Lettre>& operator<<(const SHARED_PTR(ExpressionRationnelle<Lettre> )& expression)
				{
					_expressions.push_back(expression);
					return *this;
				}
				virtual operator typename ExpressionRationnelle<Lettre>::Alphabet () const
				{	
					typename ExpressionRationnelle<Lettre>::Alphabet alphabet;
					for(auto it = _expressions.cbegin() ; it!=_expressions.cend() ; ++it)
					{
						typename ExpressionRationnelle<Lettre>::Alphabet italph = (**it);
						alphabet.insert(italph.begin(),italph.end());
					}
					return alphabet;
				}
				virtual Automate<NonDeterministe,Lettre> operator()(const typename ExpressionRationnelle<Lettre>::Alphabet& alphabet) const
				{
					Automate<NonDeterministe,Lettre> automate(alphabet);
					if(_expressions.empty())
					{
						
						automate._etats_finaux.insert(automate.ajout_etat_initial());
					}
					else
					{
						for(auto it=_expressions.cbegin() ; it!=_expressions.cend() ; ++it)
						{
							auto itauto = (**it)(alphabet);
							automate.ajout_automate(itauto);
							automate.ajout_transition_instantanee(automate._etats_finaux,itauto._etats_initiaux);
							automate._etats_finaux = itauto._etats_finaux;
							if(it==_expressions.cbegin())
							{
								automate._etats_initiaux = itauto._etats_initiaux;
							}
						}
					}
					return automate;
				}
			private:
				std::vector<SHARED_PTR(ExpressionRationnelle<Lettre> ) > _expressions;
		};

		template<typename Lettre>
		class ExpressionRationnelleAlternative : public ExpressionRationnelle<Lettre>
		{
			public:
				explicit ExpressionRationnelleAlternative() {};
				virtual ~ExpressionRationnelleAlternative() {};
				ExpressionRationnelleAlternative<Lettre>& operator<<(const SHARED_PTR(ExpressionRationnelle<Lettre> )& expression)
				{
					_expressions.push_back(expression);
					return *this;
				}
				virtual operator typename ExpressionRationnelle<Lettre>::Alphabet () const
				{	
					typename ExpressionRationnelle<Lettre>::Alphabet alphabet;
					for(auto it = _expressions.cbegin() ; it!=_expressions.cend() ; ++it)
					{
						typename ExpressionRationnelle<Lettre>::Alphabet italph = (**it);
						alphabet.insert(italph.begin(),italph.end());
					}
					return alphabet;
				}
				virtual Automate<NonDeterministe,Lettre> operator()(const typename ExpressionRationnelle<Lettre>::Alphabet& alphabet) const
				{
					Automate<NonDeterministe,Lettre> automate(alphabet);
					for(auto it=_expressions.cbegin() ; it!=_expressions.cend() ; ++it)
					{
						auto itauto = (**it)(alphabet);
						automate.ajout_automate(itauto);
						automate._etats_initiaux.insert(itauto._etats_initiaux.begin(),itauto._etats_initiaux.end());
						automate._etats_finaux.insert(itauto._etats_finaux.begin(),itauto._etats_finaux.end());
					}
					return automate;
				}
			private:
				std::vector<SHARED_PTR(ExpressionRationnelle<Lettre> ) > _expressions;
		};

	}
}

#endif
