/* 
 *  This program is an interactive FAQ to ease investigations.
 *  Copyright (C) 2012 Julien Bruguier.
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */ 

#ifndef __INVESTIGATION_DONNEES_SERVEUR__
#define __INVESTIGATION_DONNEES_SERVEUR__

#include <src/global/global.h>
namespace Investigation { namespace Donnees { DECL_SHARED_PTR(Serveur); } }
#include <src/investigation/donnees/site.h>
#include <src/investigation/donnees/html.h>

namespace Investigation
{
	namespace Donnees
	{

		class ImpossibleDePasserEnTacheDeFond : public Investigation::Global::Exception::Execution
		{
			public:
				ImpossibleDePasserEnTacheDeFond()
				:Execution(SIGTERM,"Impossible to go in background") { };
		};

		struct Serveur
		{
			static void serveur(const bool tache_de_fond, const std::string& hotes, const std::string& port, const Investigation::Donnees::SiteSP& site);
			static void requete(const Investigation::Donnees::SiteSP& site, const std::vector<std::string>& requete);
			private:
				template<bool appelle_bind,bool appelle_listen, bool appelle_connect>
					static int ouvrir(const int type, const int protocole, const std::string& ip, const std::string& port, const std::string& d);
				static void log_erreur(struct addrinfo *res, std::ostringstream& os, const std::string& fonction, int erreur);
				static bool resolution_ip_port(const struct sockaddr* sock, const socklen_t longueur, std::string& ip, std::string& port);

				class Http
				{
					public:
						template<typename oStream>
						friend oStream& operator<<(oStream& os, const Http& http)
						{
							os << "HTTP/1.1 " << http._code_retour << " " << http._type_retour << "\r\n"
								<< "Content-Type: " << http._content_type << "\r\n"
								<< "Content-Length: " << http._contenu.size() << "\r\n"
								<< "Connection: close" << "\r\n"
								<< "Cache-Control: max-age=3600" << "\r\n"
								<< "\r\n"
								<< http._contenu;
							return os;
						}
					protected:
						Http(const size_t code_retour, const std::string& type_retour, const std::string& content_type, const std::string& contenu ="")
						:_code_retour(code_retour), _type_retour(type_retour), _content_type(content_type), _contenu(contenu) {};
						Http(const size_t code_retour, const std::string& type_retour, const Investigation::Donnees::Html::Html& contenu)
						:_code_retour(code_retour), _type_retour(type_retour), _content_type("text/html; charset=UTF-8")
						{
							std::ostringstream oss;
							oss << contenu;
							_contenu = oss.str();
						};
						const size_t _code_retour;
						const std::string _type_retour;
						const std::string _content_type;
						std::string _contenu;
				};

				class HttpOK : public Http
				{
					public:
						HttpOK(const std::string& content_type, const std::string& contenu)
						:Http(200,"OK",content_type,contenu) {};
						explicit HttpOK(const Investigation::Donnees::Html::Html& contenu)
						:Http(200,"OK",contenu) {};
				};

				class HttpBadRequest : public Http
				{
					public:
						HttpBadRequest(const std::string& content_type, const std::string& contenu)
						:Http(400,"Bad Request",content_type,contenu) {};
						explicit HttpBadRequest(const Investigation::Donnees::Html::Html& contenu)
						:Http(400,"Bad Request",contenu) {};
						static HttpBadRequest defaut(const std::string& titre, const std::string& style)
						{
							Html::Html html(titre,style);
							Html::Div div("investigation error badrequest");
							Html::P p("investigation error badrequest");
							p << Html::Texte("400 Bad Request");
							div << p;
							html << div;
							return HttpBadRequest(html);
						}
				};

				class HttpNotFound : public Http
				{
					public:
						HttpNotFound(const std::string& content_type, const std::string& contenu)
						:Http(404,"Not Found",content_type,contenu) {};
						explicit HttpNotFound(const Investigation::Donnees::Html::Html& contenu)
						:Http(404,"Not Found",contenu) {};
						static HttpNotFound defaut(const std::string& titre, const std::string& style)
						{
							Html::Html html(titre,style);
							Html::Div div("investigation error notfound");
							Html::P p("investigation error notfound");
							p << Html::Texte("404 Not Found");
							div << p;
							html << div;
							return HttpNotFound(html);
						}
				};

				class HttpInternalError : public Http
				{
					public:
						HttpInternalError(const std::string& content_type, const std::string& contenu)
						:Http(500,"Internal Error",content_type,contenu) {};
						HttpInternalError(const Investigation::Donnees::Html::Html& contenu)
						:Http(500,"Internal Error",contenu) {};
						static HttpInternalError defaut(const std::string& titre, const std::string& style)
						{
							Html::Html html(titre,style);
							Html::Div div("investigation error internalerror");
							Html::P p("investigation error internalerror");
							p << Html::Texte("500 Internal Error");
							div << p;
							html << div;
							return HttpInternalError(html);
						}
				};
		};

		class OuvertureFluxImpossible : public Investigation::Global::Exception::Execution
		{
			public:
				OuvertureFluxImpossible(const std::string& descripteur, const std::string& details = std::string())
					:Execution(SIGPIPE,std::string("Impossible to open stream ")+descripteur+details) { };

		};

		class AttenteFluxImpossible : public Investigation::Global::Exception::Execution
		{               
			public:         
				AttenteFluxImpossible(const std::string& descripteur, const std::string& details = std::string())
					:Execution(SIGPIPE,std::string("Impossible to wait for connections on stream ")+descripteur+details) { };

		};              


	}
}
#endif
